#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DIM 10

/* elemento della lista di buffer */
struct bufferlistelement {
	char mem[DIM];
	struct bufferlistelement *next;
};

typedef struct bufferlistelement bufferlistelement;

/* struttura per la manipolazione della lista di buffer */
/* oltre alla radice della lista contiene il contatore degli elementi */
struct buffer {
	int count;
	bufferlistelement *root;
};

typedef struct buffer buffer;
typedef buffer *bufferlist;

/* crea una lista di buffer allocando gia' il primo elemento */
bufferlist createbufferlist(void)
{
	bufferlist q = malloc(sizeof(buffer));

	if(!q) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-1);
	}

	q->count = 1;
	if(!(q->root = malloc(sizeof(bufferlistelement)))) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-2);
	};
	q->root->next = NULL;
	return q;
}

/* inserisce un nuovo buffer in testa alla lista */
/* incrementa il contatore */
void enlargebufferlist(bufferlist bl)
{
	bufferlistelement *q;

	if(!bl)
		return;

	if(!(q = malloc(sizeof(bufferlistelement)))) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-3);
	}
	
	q->next = bl->root;
	bl->root = q;
	bl->count++;
}

/* toglie l'elemento in testa alla lista */
/* decremente il contatore */
void consumebufferlist(bufferlist bl)
{
	bufferlistelement *q;

	if(!bl)
		return;
	q = bl->root;
	bl->count--;
	bl->root = q->next;
	free(q);	
}

/* distrugge la lista di buffer */
void destroybufferlist(bufferlist bl)
{
	if(!bl)
		return; 
	while(bl->root)
		consumebufferlist(bl);
	/* dopo aver liberato tutti gli elementi della lista */
	/* distruggo anche la sua struttura per la manipolazione */
	free(bl);
}

/* restituisce l'elemento puntato dalla radice della lista di buffer */
bufferlistelement *getroot(bufferlist bl)
{
	return bl ? bl->root : NULL;
}

/* restituisce il numero di buffer contenuti nella lista (o -1 se bl e' NULL)*/
int getcount(bufferlist bl)
{
	return bl ? bl->count : -1;
}

/* dato un buffer della lista, restituisce l'indirizzo dell'array  mem */
/* in cui fisicamente sono memorizzati i caratteri letti */
char *getmem(bufferlistelement *b)
{
	return b ? b->mem : NULL;
}

/* legge una parola di lunghezza arbitraria dal file fp */
/* per parola intende una sequenza di caratteri consecutivi fra due */
/* caratteri di separazione. */
/* un carattere c e' di separazione se (*separator)(c) restituisce 1 */
char *readword(FILE *fp, char (*separator)(char))
{
	/* crea una lista di buffer */
	bufferlist buf = createbufferlist();
	/* s punta all'inizio dell'array nel primo buffer */
	char *s = getmem(getroot(buf));
	char *r,*t,*word;
	int size;

	/* legge un carattere alla volta e lo pone nel char puntato da s */
	/* fino a quando non incontra EOF o un separatore */
	while((*s = fgetc(fp)) != EOF && !(*separator)(*s))
		/* incrementa s, inoltre: */
		/* se lo spazio nel buffer corrente e' esaurito*/
		/* ne aggiunge una in testa alla lista */
		if(++s - getmem(getroot(buf)) == DIM) {
			enlargebufferlist(buf);
			/* pone s a puntare all'inizio dell'array nel nuovo buffer */
			s = getmem(getroot(buf));
		}

	/* a questo punto ho letto tutta la parola */

	/* calcolo la dimensione della parola e la assegno a size */
	size = (getcount(buf) - 1) * DIM + (s - getmem(getroot(buf))); 
	if(!size) {
		destroybufferlist(buf);
		return NULL;
	}
	/* alloco una stringa di size + 1 caratteri */
	if(!(word = malloc(size + 1))) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-4);
	}

	/* metto il carattere di terminazione */
	word[size] = 0;

	/* prelevo l'indirizzo iniziale dell'array del buffer corrente */
	r = getmem(getroot(buf));
	/* determino la posizione in word in cui copiare il contenuto del buffer corrente */
	/* e la assegno a t */
	/* NB: aritmetica dei puntatori: la seguente equivale a &word[size -s + r] */
	t = word + size - s + r;
	/* copio s - r caratteri dal buffer a t */
	strncpy(t,r,s - r);
	/* elimino il buffer appena copiato */
	consumebufferlist(buf);

	/* ciclo sugli altri buffer fino ad esaurire la lista: */
	/* Nota: questi buffer sono tutti "pieni", contengono tutti DIM caratteri della parola */
	while(getcount(buf)) {
		/* aggiorno la posizione nella stringa word, decrementando t di DIM posizioni */
		t -= DIM;
		/* copio il contenuto del buffer corrente in word, dalla posizione t */
		strncpy(t,getmem(getroot(buf)),DIM);
		/* elimino il buffer appena copiato */
		consumebufferlist(buf);
	}

	/* distruggo quanto resta della lista, cioe' la sua struttura per la manipolazione */
	destroybufferlist(buf);

	/* restituisco la parola cosi' costruita */
	return word;
}

char separator(char c)
{
	return isspace(c) || ispunct(c);
}

int main(int ac, char **av)
{
	FILE *fp;
	char *word;
	int c;

	if(ac != 2) {
		fprintf(stderr, "Uso: %s <file>\n", av[0]);
		exit(-5);
	}
	if(!(fp = fopen(av[1],"r"))) {
		fprintf(stderr, "Errore: non riesco ad aprire il file %s\n", av[1]);
		exit(-6);
	}
		
	/* ciclo fino a quando ci sono parole da leggere */	
	do {
		/* salto tutti i separatori precedenti alla prossima parola */
		while(separator(c = fgetc(fp)));
		/* rimetto nell'input l'ultimo carattere letto dato che non e' un separatore */
		ungetc(c,fp);
		/* se effettivamente c'e' una parola la leggo, la stampo, libero la memoria occupata */
		if((word = readword(fp, separator))) {
			printf("{\n%s\n}\n",word);
			free(word);
		}
	} while(word);

	fclose(fp);
	return 0;	
}
